import re
import sys
from collections import namedtuple
from .container import Well
from .unit import Unit

if sys.version_info[0] >= 3:
    xrange = range
    basestring = str

'''
    :copyright: 2016 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

'''


class ContainerType(namedtuple("ContainerType",
                    ["name", "is_tube", "well_count",
                     "well_depth_mm", "well_volume_ul",
                     "well_coating", "sterile", "cover_types",
                     "seal_types", "capabilities",
                     "shortname", "col_count", "dead_volume_ul",
                     "safe_min_volume_ul"])):

    """
    The ContainerType class holds the capabilities and properties of a
    particular container type.

    Parameters
    ----------
    name : str
      Full name describing a ContainerType.
    is_tube : bool
      Indicates whether a ContainerType is a tube (container with one well).
    well_count : int
      Number of wells a ContainerType contains.
    well_depth_mm : int
      Depth of well(s) contained in a ContainerType in millimeters.
    well_volume_ul : int
      Maximum volume of well(s) contained in a ContainerType in microliters.
    well_coating : str
      Coating of well(s) in container (ex. collagen).
    sterile : bool
      Indicates whether a ContainerType is sterile.
    cover_types: list
        List of valid covers associated with a ContainerType.
    seal_types: list
        List of valid seals associated with a ContainerType.
    capabilities : list
      List of capabilities associated with a ContainerType (ex. ["spin",
        "incubate"]).
    shortname : str
      Short name used to refer to a ContainerType.
    col_count : int
      Number of columns a ContainerType contains.
    dead_volume_ul : int
      Volume of liquid that cannot be aspirated from any given well of a
      ContainerType via liquid-handling.
    safe_min_volume_ul : int
      Minimum volume of liquid to ensure adequate volume for liquid-handling
      aspiration from any given well of a ContainerType.

    """

    def robotize(self, well_ref):
        """
        Return a robot-friendly well reference from a number of well reference
        formats.

        Example Usage:

        .. code-block:: python

            >>> p = Protocol()
            >>> my_plate = p.ref("my_plate", cont_type="6-flat", discard=True)
            >>> my_plate.robotize("A1")
            0
            >>> my_plate.robotize("5")
            5
            >>> my_plate.robotize(my_plate.well(3))
            3
            >>> my_plate.robotize(["A1", "A2"])
            [0, 1]

        Parameters
        ----------
        well_ref : str, int, Well, list[str or int or Well]
            Well reference to be robotized in string, integer or Well object
            form. Also accepts lists of str, int or Well.

        Returns
        ----------
        well_ref : int, list
            Single or list of Well references passed as rowwise integer
            (left-to-right, top-to-bottom, starting at 0 = A1).

        Raises
        ------
        TypeError
            If well reference given is not an accepted type.
        ValueError
            If well reference given exceeds container dimensions.

        """
        if isinstance(well_ref, list):
            return [self.robotize(well) for well in well_ref]

        if not isinstance(well_ref, (basestring, int, Well)):
            raise TypeError("ContainerType.robotize(): Well reference (%s) given "
                            "is not of type 'str', 'int', or 'Well'." % well_ref)

        if isinstance(well_ref, Well):
            well_ref = well_ref.index
        well_ref = str(well_ref)
        m = re.match(r"([a-z])(\d+)$", well_ref, re.I)
        if m:
            row = ord(m.group(1).upper()) - ord('A')
            col = int(m.group(2)) - 1
            well_num = row * self.col_count + col
            # Check bounds
            if row >= self.row_count():
                raise ValueError("ContainerType.robotize(): Row given exceeds "
                                 "container dimensions.")
            if col >= self.col_count or col < 0:
                raise ValueError("ContainerType.robotize(): Col given exceeds "
                                 "container dimensions.")
            return well_num
        else:
            m = re.match(r"\d+$", well_ref)
            if m:
                well_num = int(m.group(0))
                # Check bounds
                if well_num >= self.well_count or well_num < 0:
                    raise ValueError("ContainerType.robotize(): Well number "
                                     "given exceeds container dimensions.")
                return well_num
            else:
                raise ValueError("ContainerType.robotize(): Well must be in "
                                 "'A1' format or be an integer.")

    def humanize(self, well_ref):
        """
        Return the human readable form of a well index based on the well
        format of this ContainerType.

        Example Usage:

        .. code-block:: python

            >>> p = Protocol()
            >>> my_plate = p.ref("my_plate", cont_type="6-flat", discard=True)
            >>> my_plate.humanize(0)
            'A1'
            >>> my_plate.humanize(5)
            'B3'
            >>> my_plate.humanize('0')
            'A1'

        Parameters
        ----------
        well_ref : int, str, list[int or str]
            Well reference to be humanized in integer or string form.
            If string is provided, it has to be parseable into an int.
            Also accepts lists of int or str

        Returns
        -------
        well_ref : str
            Well index passed as human-readable form.

        Raises
        ------
        TypeError
            If well reference given is not an accepted type.
        ValueError
            If well reference given exceeds container dimensions.

        """
        if isinstance(well_ref, list):
            return [self.humanize(well) for well in well_ref]

        if not isinstance(well_ref, (int, basestring)):
            raise TypeError("ContainerType.humanize(): Well reference given "
                            "is not of type 'int' or 'str'.")
        try:
            well_ref = int(well_ref)
        except:
            raise TypeError("ContainerType.humanize(): Well reference given"
                            "is not parseable into 'int' format.")
        # Check bounds
        if well_ref >= self.well_count or well_ref < 0:
                raise ValueError("ContainerType.humanize(): Well reference "
                                 "given exceeds container dimensions.")
        row, col = self.decompose(well_ref)
        return "ABCDEFGHIJKLMNOPQRSTUVWXYZ"[row] + str(col + 1)

    def decompose(self, idx):
        """
        Return the (col, row) corresponding to the given well index.

        Parameters
        ----------
        well_ref : str, int
            Well index in either human-readable or integer form.

        Returns
        -------
        well_ref : tuple
            tuple containing the column number and row number of the given
            well_ref.

        """
        if not isinstance(idx, (int, basestring, Well)):
            raise TypeError("Well index given is not of type 'int' or \
                            'str'.")
        idx = self.robotize(idx)
        return (idx // self.col_count, idx % self.col_count)

    def row_count(self):
        """
        Return the number of rows of this ContainerType.

        """
        return self.well_count // self.col_count


_CONTAINER_TYPES = {
    "384-flat": ContainerType(name="384-well UV flat-bottom plate",
                              well_count=384,
                              well_depth_mm=None,
                              well_volume_ul=Unit(90.0, "microliter"),
                              well_coating=None,
                              sterile=False,
                              is_tube=False,
                              cover_types=["standard", "universal"],
                              seal_types=None,
                              capabilities=["pipette", "spin", "absorbance",
                                            "fluorescence", "luminescence",
                                            "incubate", "gel_separate",
                                            "gel_purify", "cover", "stamp",
                                            "dispense"],
                              shortname="384-flat",
                              col_count=24,
                              dead_volume_ul=Unit(5, "microliter"),
                              safe_min_volume_ul=Unit(15, "microliter")),
    "384-pcr": ContainerType(name="384-well PCR plate",
                             well_count=384,
                             well_depth_mm=None,
                             well_volume_ul=Unit(50.0, "microliter"),
                             well_coating=None,
                             sterile=None,
                             is_tube=False,
                             cover_types=["universal"],
                             seal_types=["ultra-clear", "foil"],
                             capabilities=["pipette", "spin", "thermocycle",
                                           "incubate", "gel_separate", "gel_purify",
                                           "seal", "stamp", "dispense", "cover"],
                             shortname="384-pcr",
                             col_count=24,
                             dead_volume_ul=Unit(2, "microliter"),
                             safe_min_volume_ul=Unit(3, "microliter")),
    "384-echo": ContainerType(name="384-well Echo plate",
                              well_count=384,
                              well_depth_mm=None,
                              well_volume_ul=Unit(65.0, "microliter"),
                              well_coating=None,
                              sterile=None,
                              is_tube=False,
                              cover_types=["universal"],
                              seal_types=["ultra-clear", "foil"],
                              capabilities=["pipette", "seal", "spin",
                                            "incubate", "stamp", "dispense",
                                            "cover"],
                              shortname="384-echo",
                              col_count=24,
                              dead_volume_ul=Unit(15, "microliter"),
                              safe_min_volume_ul=Unit(15, "microliter")),
    "384-flat-white-white-lv": ContainerType(name="384-well flat-bottom low volume plate",
                                             well_count=384,
                                             well_depth_mm=9.39,
                                             well_volume_ul=Unit(40.0, "microliter"),
                                             well_coating=None,
                                             sterile=False,
                                             is_tube=False,
                                             cover_types=["standard", "universal"],
                                             seal_types=None,
                                             capabilities=["absorbance", "cover", "dispense",
                                                           "fluorescence", "image_plate",
                                                           "incubate", "luminescence",
                                                           "pipette", "spin",
                                                           "stamp", "uncover"],
                                             shortname="384-flat-white-white-lv",
                                             col_count=24,
                                             dead_volume_ul=Unit(5, "microliter"),
                                             safe_min_volume_ul=Unit(15, "microliter")),
    "384-flat-white-white-tc": ContainerType(name="384-well flat-bottom low flange plate",
                                             well_count=384,
                                             well_depth_mm=11.43,
                                             well_volume_ul=Unit(80.0, "microliter"),
                                             well_coating=None,
                                             sterile=True,
                                             is_tube=False,
                                             cover_types=["standard", "universal"],
                                             seal_types=None,
                                             capabilities=["absorbance", "cover", "dispense",
                                                           "fluorescence", "image_plate",
                                                           "incubate", "luminescence",
                                                           "pipette", "spin",
                                                           "stamp", "uncover"],
                                             shortname="384-flat-white-white-tc",
                                             col_count=24,
                                             dead_volume_ul=Unit(20, "microliter"),
                                             safe_min_volume_ul=Unit(30, "microliter")),
    "384-flat-clear-clear": ContainerType(name="384-well fully clear high binding plate",
                                          well_count=384,
                                          well_depth_mm=11.43,
                                          well_volume_ul=Unit(80.0, "microliter"),
                                          well_coating="high bind",
                                          sterile=False,
                                          is_tube=False,
                                          cover_types=["standard", "universal", "low_evaporation"],
                                          seal_types=["ultra-clear", "foil"],
                                          capabilities=["incubate", "seal", "image_plate",
                                                        "stamp", "dispense", "spin",
                                                        "absorbance", "cover",
                                                        "fluorescence", "luminescence",
                                                        "pipette", "uncover"],
                                          shortname="384-flat-clear-clear",
                                          col_count=24,
                                          dead_volume_ul=Unit(5, "microliter"),
                                          safe_min_volume_ul=Unit(20, "microliter")),

    "96-flat": ContainerType(name="96-well flat-bottom plate",
                             well_count=96,
                             well_depth_mm=None,
                             well_volume_ul=Unit(340.0, "microliter"),
                             well_coating=None,
                             sterile=False,
                             is_tube=False,
                             cover_types=["standard", "universal", "low_evaporation"],
                             seal_types=None,
                             capabilities=["pipette", "spin", "absorbance",
                                           "fluorescence", "luminescence",
                                           "incubate", "gel_separate",
                                           "gel_purify", "cover", "stamp",
                                           "dispense"],
                             shortname="96-flat",
                             col_count=12,
                             dead_volume_ul=Unit(25, "microliter"),
                             safe_min_volume_ul=Unit(65, "microliter")),
    "96-flat-uv": ContainerType(name="96-well flat-bottom UV transparent \
                                plate",
                                well_count=96,
                                well_depth_mm=None,
                                well_volume_ul=Unit(340.0, "microliter"),
                                well_coating=None,
                                sterile=False,
                                is_tube=False,
                                cover_types=["standard", "universal", "low_evaporation"],
                                seal_types=None,
                                capabilities=["pipette", "spin", "absorbance",
                                              "fluorescence", "luminescence",
                                              "incubate", "gel_separate",
                                              "gel_purify", "cover", "stamp",
                                              "dispense"],
                                shortname="96-flat-uv",
                                col_count=12,
                                dead_volume_ul=Unit(25, "microliter"),
                                safe_min_volume_ul=Unit(65, "microliter")),
    "96-pcr": ContainerType(name="96-well PCR plate",
                            well_count=96,
                            well_depth_mm=None,
                            well_volume_ul=Unit(160.0, "microliter"),
                            well_coating=None,
                            sterile=None,
                            is_tube=False,
                            cover_types=["universal"],
                            seal_types=["ultra-clear", "foil"],
                            capabilities=["pipette", "sangerseq", "spin",
                                          "thermocycle", "incubate",
                                          "gel_separate", "gel_purify",
                                          "seal", "stamp", "dispense", "cover"],
                            shortname="96-pcr",
                            col_count=12,
                            dead_volume_ul=Unit(3, "microliter"),
                            safe_min_volume_ul=Unit(5, "microliter")),
    "96-deep": ContainerType(name="96-well extended capacity plate",
                             well_count=96,
                             well_depth_mm=None,
                             well_volume_ul=Unit(2000.0, "microliter"),
                             well_coating=None,
                             sterile=False,
                             cover_types=["standard"],
                             seal_types=None,
                             capabilities=["pipette", "incubate",
                                           "gel_separate", "gel_purify",
                                           "cover", "stamp", "dispense"],
                             shortname="96-deep",
                             is_tube=False,
                             col_count=12,
                             dead_volume_ul=Unit(15, "microliter"),
                             safe_min_volume_ul=Unit(30, "microliter")),
    "96-v-kf": ContainerType(name="96-well v-bottom King Fisher plate",
                             well_count=96,
                             well_depth_mm=None,
                             well_volume_ul=Unit(200.0, "microliter"),
                             well_coating=None,
                             sterile=False,
                             cover_types=["standard"],
                             seal_types=None,
                             capabilities=["pipette", "incubate",
                                           "gel_separate", "mag_dry",
                                           "mag_incubate", "mag_collect",
                                           "mag_release", "mag_mix",
                                           "cover", "stamp", "dispense"],
                             shortname="96-v-kf",
                             is_tube=False,
                             col_count=12,
                             dead_volume_ul=Unit(20, "microliter"),
                             safe_min_volume_ul=Unit(20, "microliter")),
    "96-deep-kf": ContainerType(name="96-well extended capacity \
                                King Fisher plate",
                                well_count=96,
                                well_depth_mm=None,
                                well_volume_ul=Unit(1000.0, "microliter"),
                                well_coating=None,
                                sterile=False,
                                cover_types=["universal"],
                                seal_types=None,
                                capabilities=["pipette", "incubate",
                                              "gel_separate", "mag_dry",
                                              "mag_incubate", "mag_collect",
                                              "mag_release", "mag_mix",
                                              "cover", "stamp", "dispense"],
                                shortname="96-deep-kf",
                                is_tube=False,
                                col_count=12,
                                dead_volume_ul=Unit(50, "microliter"),
                                safe_min_volume_ul=Unit(50, "microliter")),
    "24-deep": ContainerType(name="24-well extended capacity plate",
                             well_count=24,
                             well_depth_mm=None,
                             well_volume_ul=Unit(10000.0, "microliter"),
                             well_coating=None,
                             sterile=False,
                             cover_types=["universal"],
                             seal_types=None,
                             capabilities=["pipette", "incubate",
                                           "gel_separate", "gel_purify",
                                           "cover", "stamp", "dispense"],
                             shortname="24-deep",
                             is_tube=False,
                             col_count=6,
                             dead_volume_ul=Unit(15, "microliter"),
                             safe_min_volume_ul=Unit(60, "microliter")),
    "micro-2.0": ContainerType(name="2mL Microcentrifuge tube",
                               well_count=1,
                               well_depth_mm=None,
                               well_volume_ul=Unit(2000.0, "microliter"),
                               well_coating=None,
                               sterile=False,
                               cover_types=None,
                               seal_types=None,
                               capabilities=["pipette", "gel_separate",
                                             "gel_purify", "incubate", "spin"],
                               shortname="micro-2.0",
                               is_tube=True,
                               col_count=1,
                               dead_volume_ul=Unit(15, "microliter"),
                               safe_min_volume_ul=Unit(40, "microliter")),
    "micro-1.5": ContainerType(name="1.5mL Microcentrifuge tube",
                               well_count=1,
                               well_depth_mm=None,
                               well_volume_ul=Unit(1500.0, "microliter"),
                               well_coating=None,
                               sterile=False,
                               cover_types=None,
                               seal_types=None,
                               capabilities=["pipette", "gel_separate",
                                             "gel_purify", "incubate", "spin"],
                               shortname="micro-1.5",
                               is_tube=True,
                               col_count=1,
                               dead_volume_ul=Unit(15, "microliter"),
                               safe_min_volume_ul=Unit(20, "microliter")),
    "6-flat": ContainerType(name="6-well cell culture plate",
                            well_count=6,
                            well_depth_mm=None,
                            well_volume_ul=Unit(5000.0, "microliter"),
                            well_coating=None,
                            sterile=False,
                            cover_types=["standard", "universal"],
                            seal_types=None,
                            capabilities=["cover", "incubate", "image_plate"],
                            shortname="6-flat",
                            is_tube=False,
                            col_count=3,
                            dead_volume_ul=Unit(400, "microliter"),
                            safe_min_volume_ul=Unit(600, "microliter")),
    "1-flat": ContainerType(name="1-well flat-bottom plate",
                            well_count=1,
                            well_depth_mm=None,
                            well_volume_ul=Unit(80000.0, "microliter"),
                            well_coating=None,
                            sterile=False,
                            cover_types=["universal"],
                            seal_types=None,
                            capabilities=["cover", "incubate"],
                            shortname="1-flat",
                            is_tube=False,
                            col_count=1,
                            dead_volume_ul=Unit(36000, "microliter"),
                            safe_min_volume_ul=Unit(40000, "microliter")),
}



from __future__ import division, print_function
from pint import UnitRegistry
from pint.quantity import _Quantity
import sys

if sys.version_info[0] >= 3:
    string_type = str
else:
    string_type = basestring

'''
    :copyright: 2016 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

'''

# Preload UnitRegistry (Use default Pints definition file as a base)
_UnitRegistry = UnitRegistry()

'''Map string representation of Pint units over to Autoprotocol format'''
# Map Temperature Unit names
_UnitRegistry._units["degC"]._name = "celsius"
_UnitRegistry._units["celsius"]._name = "celsius"
_UnitRegistry._units["degF"]._name = "fahrenheit"
_UnitRegistry._units["fahrenheit"]._name = "fahrenheit"
_UnitRegistry._units["degR"]._name = "rankine"
_UnitRegistry._units["rankine"]._name = "rankine"
# Map Speed Unit names
_UnitRegistry._units["revolutions_per_minute"]._name = "rpm"

'''Add support for Molarity Unit'''
_UnitRegistry.define('molar = mole/liter = M')


class UnitError(Exception):
    """
    Exceptions from creating new Unit instances with bad inputs.
    """

    message_text = "Unit error for %s"

    def __init__(self, value):
        super(UnitError, self).__init__(self.message_text % value)
        self.value = value


class UnitStringError(UnitError):
    message_text = ("Invalid format for %s: when building a Unit from a "
                    "string argument, string must be in \'1:meter\' format.")


class UnitValueError(UnitError):
    message_text = ("Invalid value for %s: when building a Unit from a "
                    "value argument, value must be numeric.")


class Unit(_Quantity):
    """
    A representation of a measure of physical quantities such as length,
    mass, time and volume.
    Uses Pint's Quantity as a base class for implementing units and
    inherits functionalities such as conversions and proper unit
    arithmetic.
    Note that the magnitude is stored as a double-precision float, so
    there are inherent issues when dealing with extremely large/small
    numbers as well as numerical rounding for non-base 2 numbers.

    Example
    -------

        .. code-block:: python

            vol_1 = Unit(10, 'microliter')
            vol_2 = Unit(10, 'liter')
            print(vol_1 + vol_2)

            time_1 = Unit(1, 'second')
            speed_1 = vol_1/time_1
            print (speed_1)
            print (speed_1.to('liter/hour'))

    Returns
    -------

        .. code-block:: json

            10000010.0:microliter
            10.0:microliter / second
            0.036:liter / hour

    """
    def __new__(cls, value, units=None):
        cls._REGISTRY = _UnitRegistry
        cls.force_ndarray = False

        # Automatically return Unit if Unit is provided
        if isinstance(value, Unit):
            return value

        # Automatically parse String if no units provided
        if not units and isinstance(value, string_type):
            try:
                value, units = value.split(":")
            except ValueError:
                raise UnitStringError(value)
        try:
            return super(Unit, cls).__new__(cls, float(value), units)
        except ValueError:
            raise UnitValueError(value)

    def __init__(self, value, units=None):
        super(Unit, self).__init__()
        self.unit = self.units.__str__()

    @staticmethod
    def fromstring(s):
        """
        Convert a string representation of a unit into a Unit object.

        Example
        -------

        .. code-block:: python

            Unit.fromstring("10:microliter")

        becomes

        .. code-block:: python

            Unit(10, "microliter")

        Parameters
        ----------
        s : str
            String in the format of "value:unit"

        """
        if isinstance(s, Unit):
            return s
        else:
            return Unit(s)

    def __str__(self):
        return ":".join([str(self._magnitude), "^".join(self.unit.split("**"))]).replace(" ", "")

    def __repr__(self):
        return "Unit({0}, '{1}')".format(self._magnitude, self._units)

    def _mul_div(self, other, magnitude_op, units_op=None):
        '''
        Extends Pint's base _Quantity multiplication/division
        implementation by checking for dimensionality
        '''
        if isinstance(other, Unit):
            if self.dimensionality == other.dimensionality:
                other = other.to(self.unit)
        return super(Unit, self)._mul_div(other, magnitude_op, units_op)

    def _imul_div(self, other, magnitude_op, units_op=None):
        '''
        Extends Pint's base _Quantity multiplication/division
        implementation by checking for dimensionality
        '''
        if isinstance(other, Unit):
            if self.dimensionality == other.dimensionality:
                other = other.to(self.unit)
        return super(Unit, self)._imul_div(other, magnitude_op, units_op)



#!/usr/bin/env python
from setuptools import setup

setup(
    name='autoprotocol',
    url='http://github.com/autoprotocol/autoprotocol-python',
    author='Vanessa Biggers',
    description='Python library for generating Autoprotocol',
    author_email="vanessa@transcriptic.com",
    version='3.7.2',
    test_suite='test',
    install_requires=[
        'Pint>=0.7.2'
    ],
    tests_require=[
        'coverage>=4.0.3',
        'tox>=2.3.1'
    ],
    packages=['autoprotocol']
)



from .unit import Unit
from textwrap import dedent


'''
    :copyright: 2016 by The Autoprotocol Development Team, see AUTHORS
        for more details.
    :license: BSD, see LICENSE for more details

'''


def quad_ind_to_num(q):
    """
    Convert a 384-well plate quadrant well index into its corresponding
    integer form.

    "A1" -> 0
    "A2" -> 1
    "B1" -> 2
    "B2" -> 3

    Parameters
    ----------
    q : int, str
        A string or integer representing a well index that corresponds to a
        quadrant on a 384-well plate.

    """
    if isinstance(q, str):
        q = q.lower()
    if q in ["a1", 0]:
        return 0
    elif q in ["a2", 1]:
        return 1
    elif q in ["b1", 24]:
        return 2
    elif q in ["b2", 25]:
        return 3
    else:
        raise ValueError("Invalid quadrant index.")


def quad_num_to_ind(q, human=False):
    """
    Convert a 384-well plate quadrant integer into its corresponding well index.

    0 -> "A1" or 0
    1 -> "A2" or 1
    2 -> "B1" or 24
    3 -> "B2" or 25

    Parameters
    ----------
    q : int
        An integer representing a quadrant number of a 384-well plate.
    human : bool, optional
        Return the corresponding well index in human readable form instead of
        as an integer if True.

    """
    if q == 0:
        if human:
            return "A1"
        else:
            return 0
    elif q == 1:
        if human:
            return "A2"
        else:
            return 1
    elif q == 2:
        if human:
            return "B1"
        else:
            return 24
    elif q == 3:
        if human:
            return "B2"
        else:
            return 25
    else:
        raise ValueError("Invalid quadrant number.")


def check_valid_origin(origin, stamp_type, columns, rows):
    # Checks if selected well is a valid origin destination for the given plate
    # Assumption: SBS formatted plates and 96-tip layout
    robotized_origin = origin.index
    well_count = origin.container.container_type.well_count
    col_count = origin.container.container_type.col_count
    row_count = well_count // col_count

    if well_count == 96:
        if stamp_type == "full":
            if robotized_origin != 0:
                raise ValueError("For full 96-well transfers, origin has "
                                 "to be well 0.")
        elif stamp_type == "row":
            if (robotized_origin % col_count) != 0 or robotized_origin > ((
                    row_count - rows) * col_count):
                raise ValueError("For row transfers, origin "
                                 "has to be specified within the left "
                                 "column and not more than allowed by shape.")
        else:
            if robotized_origin > (col_count - columns) or robotized_origin < 0:
                raise ValueError("For column transfers, origin "
                                 "has to be specified within the top "
                                 "column and not more than allowed by shape.")
    elif well_count == 384:
        if stamp_type == "full":
            if robotized_origin not in [0, 1, 24, 25]:
                raise ValueError("For full 384-well transfers, origin has "
                                 "to be well 0, 1, 24 or 25.")
        elif stamp_type == "row":
            if (robotized_origin % col_count) not in [0, 1] or (
                robotized_origin >= ((row_count - ((
                    rows - 1) * 2)) * col_count)):
                raise ValueError("For row transfers, origin"
                                 "has to be specified within the left "
                                 "two columns and not more than allowed by "
                                 "shape.")
        else:
            if robotized_origin >= col_count * 2 or (
                robotized_origin < 0) or (
                robotized_origin % col_count >= (
                    col_count - ((columns - 1) * 2))):
                raise ValueError("For column transfers, origin "
                                 "has to be specified within the top "
                                 "two columns and not more than allowed by "
                                 "shape.")
    else:
        raise RuntimeError("Unsupported plate type for checking origin.")


def check_stamp_append(current_xfer, prev_xfer_list, maxTransfers=3,
                       maxContainers=3,
                       volumeSwitch=Unit.fromstring("31:microliter")):
    """
    Checks whether current stamp can be appended to previous stamp instruction.
    """
    prev_cols = prev_xfer_list[0]["shape"]["columns"]
    prev_rows = prev_xfer_list[0]["shape"]["rows"]
    cols = current_xfer["shape"]["columns"]
    rows = current_xfer["shape"]["rows"]

    # Ensure Instruction contains either all full plate or selective (all rows
    # or all columns)
    if (prev_cols == cols == 12) and (prev_rows == rows == 8):
        axis_key = None
    elif prev_cols == 12:
        axis_key = "rows"
        if cols != 12:
            return False
    elif prev_rows == 8:
        axis_key = "columns"
        if rows != 8:
            return False

    # Ensure Instruction contain the same volume type as defined by TCLE
    # Currently volumeSwitch is hardcoded to check against the two tip volume
    # types used in TCLE
    if prev_xfer_list[0]["transfer"][0]["volume"] <= volumeSwitch:
        if current_xfer["transfer"][0]["volume"] > volumeSwitch:
            return False
    elif prev_xfer_list[0]["transfer"][0]["volume"] > volumeSwitch:
        if current_xfer["transfer"][0]["volume"] <= volumeSwitch:
            return False

    # Check if maximum Transfers/Containers is reached
    originList = ([y["from"] for x in prev_xfer_list for y in x["transfer"]] +
                  [y["to"] for x in prev_xfer_list for y in x["transfer"]] +
                  [y["from"] for y in current_xfer["transfer"]] +
                  [y["to"] for y in current_xfer["transfer"]])

    if axis_key:
        num_prev_xfers = sum([x["shape"][axis_key] for x in prev_xfer_list])
        num_current_xfers = current_xfer["shape"][axis_key]
    else:
        num_prev_xfers = len(prev_xfer_list)
        num_current_xfers = 1

    if (num_prev_xfers + num_current_xfers > maxTransfers or
            len(set(map(lambda x: x.container, originList))) > maxContainers):
        return False

    return True


def check_valid_mag(container, head):
    shortname = container.container_type.shortname

    if head == "96-deep":
        if shortname not in ["96-v-kf", "96-deep-kf", "96-deep"]:
            raise ValueError("{} container is not compatible with {} head"
                             .format(container.container_type.shortname, head))
    elif head == "96-pcr":
        if shortname not in ["96-pcr", "96-v-kf", "96-flat", "96-flat-uv"]:
            raise ValueError("{} container is not compatible with {} head"
                             .format(container.container_type.shortname, head))


def check_valid_mag_params(mag_dict):
    if "frequency" in mag_dict:
        if Unit.fromstring(mag_dict["frequency"]) < Unit.fromstring("0:hertz"):
            raise ValueError("Frequency set at {}, must not be less than 0:hertz"
                             .format(mag_dict["frequency"]))

    if "temperature" in mag_dict and mag_dict["temperature"]:
        if Unit.fromstring(mag_dict["temperature"]) < Unit.fromstring("-273.15:celsius"):
            raise ValueError("Temperature set at {}, must not be less than absolute zero'"
                             .format(mag_dict["temperature"]))
    elif "temperature" in mag_dict and not mag_dict["temperature"]:
            del mag_dict["temperature"]

    if "amplitude" in mag_dict:
        if mag_dict["amplitude"] > mag_dict["center"]:
            raise ValueError("'amplitude': {}, must be less than 'center': {}"
                             .format(mag_dict["amplitude"], mag_dict["center"]))
        if mag_dict["amplitude"] < 0:
            raise ValueError("Amplitude set at %s, must not be negative" % mag_dict["amplitude"])

    if any(kw in mag_dict for kw in ("center", "bottom_position", "tip_position")):
        tip_position = mag_dict.get("tip_position")
        bottom_position = mag_dict.get("bottom_position", tip_position)
        position = mag_dict.get("center", bottom_position)

        if position < 0:
            raise ValueError("Tip head position set at %s, must not be negative" % position)

    if "magnetize" in mag_dict:
        if not isinstance(mag_dict["magnetize"], bool):
            raise ValueError("Magnetize set at: %s, must be boolean" % mag_dict["magnetize"])


def check_valid_gel_purify_band(band):
    from .container import Well

    if not isinstance(band, dict):
        error_str = dedent("""
            Bands for gel_purify must be a dictionary in the form of
                {
                  'elution_volume': volume,
                  'elution_buffer': str,
                  'band_size_range': {
                    'min_bp': int,
                    'max_bp': int
                    },
                  'destination': well
                }
        """)
        raise AttributeError(error_str)

    if not all(k in band.keys() for k in ["elution_buffer", "band_size_range", "destination", "elution_volume"]):
        raise KeyError("Band parameter keys must be: 'elution_buffer', 'band_size_range', 'destination' 'elution_volume'.")
    if not isinstance(band["elution_volume"], Unit):
        raise ValueError("All band elution volumes must be of type Unit.")
    if band["elution_volume"] <= Unit(0, "microliter"):
        raise ValueError("Band elution volume: %s must be greater than 0:microliter." % band["elution_volume"])
    if not isinstance(band["destination"], Well):
        raise ValueError("All band destinations must be of type Well.")
    if not isinstance(band["band_size_range"], dict):
        raise AttributeError("Band parameter 'band_size_range' must be a dict with keys: 'max_bp', 'min_bp.")
    if not all(k in band["band_size_range"].keys() for k in ["min_bp", "max_bp"]):
        raise KeyError("Band parameter 'band_size_range' keys must be: 'max_bp', 'min_bp'.")
    if not band["band_size_range"]["max_bp"] > band["band_size_range"]["min_bp"]:
        raise ValueError("max_bp must be greater than min_bp")


def check_valid_gel_purify_extract(extract):

    from .container import Well

    if not isinstance(extract, dict):
        raise AttributeError("Extract for gel_purify must be a dictionary in the form of {'source': well, 'band_list': list, 'gel': int, 'lane': int}")
    if not all(k in extract.keys() for k in ["source", "band_list", "lane", "gel"]):
        raise KeyError("Extract parameter keys must be 'source', 'band_list', 'lane', 'gel'.")
    if not isinstance(extract["source"], Well):
        raise ValueError("All extract sources must be of type Well.")

    for band in extract["band_list"]:
        check_valid_gel_purify_band(band)


def make_gel_extract_params(source, band_list, lane=None, gel=None):

    if isinstance(band_list, dict):
        band_list = [band_list]

    for band in band_list:
        check_valid_gel_purify_band(band)

    extract = {
        "source": source,
        "band_list": band_list,
        "lane": lane,
        "gel": gel
    }

    check_valid_gel_purify_extract(extract)

    return extract


def make_band_param(elution_buffer, elution_volume, max_bp, min_bp, destination):

    band = {
        "band_size_range": {
            "min_bp": min_bp, "max_bp": max_bp
        },
        "elution_volume": Unit(elution_volume),
        "elution_buffer": elution_buffer,
        "destination": destination
    }

    check_valid_gel_purify_band(band)

    return band


class make_dottable_dict(dict):

    '''Enable dictionaries to be accessed using dot notation instead of bracket
    notation.  This class should probably never be used.

    Example
    -------
    .. code-block:: python

        >>> d = {"foo": {
                    "bar": {
                        "bat": "Hello!"
                        }
                    }
                }

        >>> print d["foo"]["bar"]["bat"]
        Hello!

        >>> d = make_dottable_dict(d)

        >>> print d.foo.bar.bat
        Hello!

    Parameters
    ----------
    dict : dict
        Dictionary to be made dottable.

    '''

    def __getattr__(self, attr):
        if type(self[attr]) == dict:
            return make_dottable_dict(self[attr])
        return self[attr]

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def deep_merge_params(defaults, override):
    """Merge two dictionaries while retaining common key-value pairs.

    Parameters
    ----------
    defaults : dict
        Default dictionary to compare with overrides.
    override : dict
        Dictionary containing additional keys and/or values to override those
        corresponding to keys in the defaults dicitonary.

    """
    defaults = make_dottable_dict(defaults.copy())
    for key, value in override.items():
        if isinstance(value, dict):
            # get node or create one
            defaults[key] = deep_merge_params(defaults.get(key, {}), value)
        else:
            defaults[key] = value

    return defaults


def incubate_params(duration, shake_amplitude=None, shake_orbital=None):
    """
    Create a dictionary with incubation parameters which can be used as input
    for instructions. Currenly supports plate reader instructions and could be
    extended for use with other instructions.

    Parameters
    ----------
    shake_amplitude: str, Unit
        amplitude of shaking between 1 and 6:millimeter
    shake_orbital: bool
        True for oribital and False for linear shaking
    duration: str, Unit
        time for shaking
    """

    incubate_dict = {}
    incubate_dict["duration"] = duration
    if (shake_amplitude is not None) and (shake_orbital is not None):
        shake_dict = {
            "amplitude": shake_amplitude,
            "orbital": shake_orbital
        }
        incubate_dict["shaking"] = shake_dict
    elif (shake_amplitude is not None) ^ (shake_orbital is not None):
        raise RuntimeError("Both `shake_amplitude`: {} and `shake_orbital`: {} "
                           "must not be None for shaking to be set".format(shake_amplitude, shake_orbital))

    check_valid_incubate_params(incubate_dict)

    return incubate_dict


def check_valid_incubate_params(idict):
    """Check to be sure incubate_params are structured correctly

    .. code-block:: json

        {
          "shaking": {
            "amplitude": str, Unit
            "orbital": bool
            }
        "duration": str, Unit
        }

    where duration is required, shaking is optional, and amplitude and orbital
    are both required if shaking is specified.

    """

    if 'duration' not in idict:
        raise RuntimeError("For the incubation dictionary: %s, `duration` must be specified" % idict)
    else:
        if Unit.fromstring(idict['duration']) <= Unit("0:second"):
            raise ValueError("duration: %s must be positive" % idict['duration'])

    if "shaking" in idict:
        shaking = idict["shaking"]
        if "orbital" in shaking and "amplitude" in shaking:
            if not isinstance(shaking["orbital"], bool):
                raise ValueError("shake_orbital: %s must be a boolean" % shaking["orbital"])
            if Unit.fromstring(shaking["amplitude"]) < Unit.fromstring("0:millimeter"):
                raise ValueError("shake_amplitude: %s must be positive" % shaking["amplitude"])
        else:
            raise RuntimeError("Both `shake_amplitude`: {} and `shake_orbital`: "
                               "{} must not be None for shaking to be set"
                               .format(shaking.get("amplitude"), shaking.get("orbital")))

    return True


def is_valid_well(param):
    """Checks if an input is of type Well, Wellgroup or list of type Well.

    Example Usage:

    .. code-block:: python

        if not is_valid_well(source):
            raise TypeError("Source must be of type Well, list of Wells, or "
                            "WellGroup.")

    Parameters
    ----------
    param :
        Parameter to validate is type Well, WellGroup, list of Wells.

    Returns
    -------
    bool :
        Returns True if param is of type Well, WellGroup or list of type Well.
    """

    from autoprotocol.container import Well, WellGroup
    if not isinstance(param, (Well, WellGroup, list)):
        return False
    if isinstance(param, list):
        if not all(isinstance(well, Well) for well in param):
            return False
    return True
